#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <poll.h>
#include <signal.h>
#include <stdarg.h>

static void die(const char *fmt, ...) __attribute__((format(printf, 1, 2)));
static void die(const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	_exit(1);
}

int main(int argc, char *argv[])
{
	char buffer[128] = { };
	const unsigned int stamp = time(NULL);
	const int sock_fd = socket(PF_INET, SOCK_STREAM, 0);
	const pid_t self_pid = getpid();
	char *vmlinux = NULL;
	int vmcore_fd = EOF;
	int monitor_fd = EOF;
	int pipe_fd[2] = { EOF, EOF };
	signal(SIGCLD, SIG_IGN);
	if (argc < 3)
		die("Usage: %s /path/to/vmlinux /path/to/qemu arguments\n",
		    argv[0]);
	vmlinux = argv[1];
	if (access(vmlinux, R_OK))
		die("Invalid vmlinux location: %s\n", vmlinux);
	if (pipe(pipe_fd))
		die("Can't create pipe for vmcore image.\n");
	snprintf(buffer, sizeof(buffer) - 1, "/var/tmp/vmcore.%u.%u", stamp,
		 self_pid);
	vmcore_fd = open(buffer, O_RDWR | O_CREAT | O_EXCL, 0600);
	if (vmcore_fd == EOF || unlink(buffer))
		die("Can't create temporary file for vmcore image.\n");
	/* Open a tcp socket for the monitor interface. */
	{
		struct sockaddr_in addr = {
			.sin_family = AF_INET,
			.sin_addr.s_addr = htonl(INADDR_LOOPBACK)
		};
		socklen_t size = sizeof(addr);
		if (bind(sock_fd, (struct sockaddr *) &addr, sizeof(addr)) ||
		    listen(sock_fd, 1) ||
		    getsockname(sock_fd, (struct sockaddr *) &addr, &size) ||
		    fcntl(sock_fd, F_SETFL,
			  O_NONBLOCK | fcntl(sock_fd, F_GETFL)))
			die("Can't create socket for monitor.\n");
		snprintf(buffer, sizeof(buffer) - 1,
			 "tcp:127.0.0.1:%u,nodelay", ntohs(addr.sin_port));
	}
	/* Start a qemu process. */
	switch(fork()) {
		int i;
		char **args;
	case 0:
		argv += 2;
		argc -= 2;
		args = calloc(sizeof(char *), argc + 8);
		if (!args)
			_exit(1);
		for (i = 0; i < argc; i++)
			args[i] = argv[i];
		args[i++] = "-monitor";
		args[i++] = buffer;
		args[i++] = "-device";
		args[i++] = "pvpanic";
		if (close(pipe_fd[0]) == 0 && close(pipe_fd[1]) == 0 &&
		    close(sock_fd) == 0 && close(vmcore_fd) == 0)
			execvp(args[0], args);
		_exit(1);
	case EOF:
		die("Can't find qemu process.\n");
	}
	/* Wait for the qemu process to start. */
	{
		struct pollfd pfd = { .fd = sock_fd, .events = POLLIN };
		poll(&pfd, 1, 30 * 1000);
		monitor_fd = accept(sock_fd, NULL, 0);
		if (close(sock_fd) || monitor_fd == EOF)
			die("Can't find qemu process.\n");
		pfd.fd = monitor_fd;
		while (1) {
			char c;
			if (poll(&pfd, 1, 30 * 1000) != 1 ||
			    read(monitor_fd, &c, 1) != 1 || c == ')')
				break;
		}
	}
	/* Wait until pvpanic reports the kernel panic event. */
	while (1) {
		struct pollfd pfd = { .fd = monitor_fd, .events = POLLIN };
		int i;
		snprintf(buffer, sizeof(buffer) - 1, "info status\n");
		write(monitor_fd, buffer, strlen(buffer));
		memset(buffer, 0, sizeof(buffer));
		for (i = 0; i < sizeof(buffer) - 1; i++) {
			if (poll(&pfd, 1, 30 * 1000) != 1 ||
			    read(monitor_fd, buffer + i, 1) != 1 ||
			    buffer[i] == ')')
				break;
			/* Echo back contains escape sequences. */
			if (buffer[i] == 0x1B)
				i = -1;
		}
		if (buffer[i] != ')')
			return 0;
		if (strstr(buffer, "(guest-panicked)"))
			break;
		sleep(10);
	}
	/* Receive vmcore image from the qemu process. */
	{
		struct pollfd pfd = { .fd = pipe_fd[0], .events = POLLIN };
		snprintf(buffer, sizeof(buffer) - 1,
			 "dump-guest-memory /proc/%u/fd/%u\n", self_pid,
			 pipe_fd[1]);
		write(monitor_fd, buffer, strlen(buffer));
		/* Close the writer side after the qemu process opened it. */
		poll(&pfd, 1, 30 * 1000);
		if (close(pipe_fd[1]))
			die("Can't close pipe.\n");
		while (splice(pipe_fd[0], NULL, vmcore_fd, NULL,
			      1024 * 1048576, 0) > 0);
		if (close(pipe_fd[0]))
			die("Can't close pipe.\n");
	}
	/* Terminate the qemu process. */
	{
		snprintf(buffer, sizeof(buffer) - 1, "q\n");
		write(monitor_fd, buffer, strlen(buffer));
		if (close(monitor_fd))
			die("Can't close monitor.\n");
	}
	/*
	 * Start the crash utility for automatic analysys.
	 * The vmcore file is deleted automatically when the process ends.
	 */
	{
		/* TODO: Redirect stdin to a script containing commands. */
		/* TODO: Redirect stdout to a location to save the output. */
		snprintf(buffer, sizeof(buffer) - 1, "/proc/%u/fd/%u",
			 self_pid, vmcore_fd);
		execlp("crash", "crash", "--", vmlinux, buffer, NULL);
	}
	return 1;
}
